#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json


def job_human_readable(job):
    """
    Generates a human readable string for the given job.
    """

    s = (
        """Request ID: %(job_id)s
Status: %(job_state)s
Client DN: %(user_dn)s
Reason: %(reason)s
Submission time: %(submit_time)s
Priority: %(priority)d
VO Name: %(vo_name)s
"""
        % job
    )

    return s


def job_list_human_readable(job_list):
    """
    Generates a guman readable string for the given job list.
    """
    jobstr = []
    for job in job_list:
        jobstr.append(job_human_readable(job))
    return "\n".join(jobstr)


def job_list_as_json(job_list):
    """
    Serializes a job list into JSON
    """
    return json.dumps(job_list, indent=2)


def job_as_json(job):
    """
    Serializes a job into JSON
    """
    return json.dumps(job, indent=2)



#   Copyright 2014-2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
from urllib.parse import quote


class Ban:
    def __init__(self, context):
        self.context = context

    def ban_dn(self, dn):
        answer = self.context.post_json("/ban/dn", {"user_dn": dn})
        return json.loads(answer)

    def ban_se(self, se, status, timeout, allow_submit):
        answer = self.context.post_json(
            "/ban/se",
            {
                "storage": se,
                "status": status,
                "timeout": timeout,
                "allow_submit": allow_submit,
            },
        )
        return json.loads(answer)

    def unban_dn(self, dn):
        self.context.delete("/ban/dn?user_dn=%s" % quote(dn, ""))

    def unban_se(self, se):
        self.context.delete("/ban/se?storage=%s" % quote(se, ""))



#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from .base import Base
from fts3.rest.client import Submitter


class JobCanceller(Base):
    def __init__(self):
        super(JobCanceller, self).__init__(
            description="""
            This command can be used to cancel a running job.  It returns the final state of the canceled job.
            Please, mind that if the job is already in a final state (FINISHEDDIRTY, FINISHED, FAILED),
            this command will return this state.
            You can additionally cancel only a subset appending a comma-separated list of file ids
            """,
            example="""
            $ %(prog)s -s https://fts3-devel.cern.ch:8446 c079a636-c363-11e3-b7e5-02163e009f5a
            FINISHED
            $ %(prog)s -s https://fts3-devel.cern.ch:8446 c079a636-c363-11e3-b7e5-02163e009f5a:5658,5659,5670
            CANCELED, CANCELED, CANCELED
            """,
        )

    def run(self):
        if ":" in self.args[0]:
            job_id, file_ids = self.args[0].split(":")
            file_ids = file_ids.split(",")
        else:
            job_id = self.args[0]
            file_ids = None

        context = self._create_context()
        submitter = Submitter(context)
        result = submitter.cancel(job_id, file_ids)
        if file_ids:
            if isinstance(result, str):
                self.logger.info(result)
            else:
                self.logger.info("\n".join(result))
        else:
            self.logger.info(result["job_state"])



#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from configparser import ConfigParser
from optparse import OptionParser, IndentedHelpFormatter
import logging
import os
import socket
import sys

from fts3.rest.client import Context
from fts3 import __version__ as CLIENT_VERSION


CONFIG_FILENAMES = ["/etc/fts3/fts3client.cfg", os.path.expanduser("~/.fts3client.cfg")]

CONFIG_DEFAULTSECTION = "Main"
CONFIG_DEFAULTS = {
    "verbose": "false",
    "endpoint": "None",
    "json": "false",
    "ukey": "None",
    "ucert": "None",
}


class _Formatter(IndentedHelpFormatter):
    def format_epilog(self, epilog):
        if not epilog:
            return ""
        else:
            lines = ["Example:"]
            indent = (self.current_indent + self.indent_increment) * " "
            for l in epilog.splitlines():
                nl = l.strip() % {"prog": sys.argv[0]}
                if len(nl) > 0:
                    lines.append(indent + nl)
            return "\n" + "\n".join(lines) + "\n\n"


def _get_local_endpoint():
    """
    Generate an endpoint using the machine hostname
    """
    return "https://%s:8446" % socket.getfqdn()


class Base:
    def __init__(self, extra_args=None, description=None, example=None):
        self.logger = logging.getLogger("fts3")

        # Common CLI options
        usage = None
        if extra_args:
            usage = "usage: %prog [options] " + extra_args

        config = ConfigParser(defaults=CONFIG_DEFAULTS)

        section = CONFIG_DEFAULTSECTION
        config.read(CONFIG_FILENAMES)

        # manually set the section in edge cases
        if not config.has_section("Main"):
            section = "DEFAULT"

        # manually get values for which we need to support None
        opt_endpoint = config.get(section, "endpoint")
        if opt_endpoint == "None":
            opt_endpoint = None
        opt_ukey = config.get(section, "ukey")
        if opt_ukey == "None":
            opt_ukey = None
        opt_ucert = config.get(section, "ucert")
        if opt_ucert == "None":
            opt_ucert = None

        self.opt_parser = OptionParser(
            usage=usage, description=description, epilog=example, formatter=_Formatter()
        )

        self.opt_parser.add_option(
            "-v",
            "--verbose",
            dest="verbose",
            action="store_true",
            help="verbose output.",
            default=config.getboolean(section, "verbose"),
        )
        self.opt_parser.add_option(
            "-s",
            "--endpoint",
            dest="endpoint",
            help="FTS3 REST endpoint.",
            default=opt_endpoint,
        )
        self.opt_parser.add_option(
            "-j",
            "--json",
            dest="json",
            action="store_true",
            help="print the output in JSON format.",
            default=config.getboolean(section, "json"),
        )
        self.opt_parser.add_option(
            "--key",
            dest="ukey",
            help="the user certificate private key.",
            default=opt_ukey,
        )
        self.opt_parser.add_option(
            "--cert", dest="ucert", help="the user certificate.", default=opt_ucert
        )
        self.opt_parser.add_option(
            "--capath",
            dest="capath",
            default="/etc/grid-security/certificates",
            help="use the specified directory to verify the peer",
        )
        self.opt_parser.add_option(
            "--insecure",
            dest="verify",
            default=True,
            action="store_false",
            help="do not validate the server certificate",
        )
        self.opt_parser.add_option(
            "--access-token",
            dest="access_token",
            help="OAuth2 access token (supported only by some endpoints, takes precedence)",
            default=None,
        )

    def __call__(self, argv=sys.argv[1:]):
        (self.options, self.args) = self.opt_parser.parse_args(argv)
        if self.options.endpoint is None:
            self.options.endpoint = _get_local_endpoint()
        if self.options.verbose:
            self.logger.setLevel(logging.DEBUG)
        self.validate()
        return self.run()

    def validate(self):
        """
        Should be implemented by inheriting classes to validate the command line arguments.
        The implementation is assumed to call sys.exit() to abort if needed
        """
        pass

    def run(self):
        """
        Implementation of the command
        """
        raise NotImplementedError(
            "Run method not implemented in %s" % type(self).__name__
        )

    def _create_context(self):
        user_agent = "fts-rest-cli/" + CLIENT_VERSION
        return Context(
            self.options.endpoint,
            ukey=self.options.ukey,
            ucert=self.options.ucert,
            verify=self.options.verify,
            access_token=self.options.access_token,
            capath=self.options.capath,
            user_agent=user_agent,
        )



#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json

from .base import Base
from fts3.rest.client import Inquirer


class WhoAmI(Base):
    def __init__(self):
        super(WhoAmI, self).__init__(
            description="""
            This command exists for convenience. It can be used to check, as the name suggests,
            who are we for the server.
            """,
            example="""
            $ %(prog)s -s https://fts3-pilot.cern.ch:8446
            User DN: /DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=saketag/CN=678984/CN=Alejandro Alvarez Ayllon
            VO: dteam
            VO: dteam/cern
            Delegation id: 9a4257f435fa2010
            """,
        )

    def run(self):
        context = self._create_context()
        inquirer = Inquirer(context)
        whoami = inquirer.whoami()

        if self.options.json:
            print(json.dumps(whoami, indent=2))
        else:
            self.logger.info("User DN: %s" % whoami["dn"][0])
            for vo in whoami["vos"]:
                self.logger.info("VO: %s" % vo)
            for vo_id in whoami["vos_id"]:
                self.logger.info("VO id: %s" % vo_id)
            self.logger.info("Delegation id: %s" % whoami["delegation_id"])
            self.logger.info("Base id: %s" % whoami["base_id"])



#   Copyright 2014-2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import timedelta
import json
import logging
import sys
import time

from .base import Base
from fts3.rest.client import Submitter, Delegator, Inquirer


def _metadata(data):
    try:
        return json.loads(data)
    except Exception:
        return str(data)


class JobDeletionSubmitter(Base):
    def __init__(self):
        super(JobDeletionSubmitter, self).__init__(
            extra_args="SURL1 [SURL2] [SURL3] [...]",
            description="""
            This command can be used to submit a deletion job to FTS3. It supports simple and bulk submissions.
            """,
            example="""
            $ %(prog)s -s https://fts3-devel.cern.ch:8446 gsiftp://source.host/file1 gsiftp://source.host/file2
            Job successfully submitted.
            Job id: 9fee8c1e-c46d-11e3-8299-02163e00a17a

            $ %(prog)s -s https://fts3-devel.cern.ch:8446 -f bulk.list
            Job successfully submitted.
            Job id: 9fee8c1e-c46d-11e3-8299-02163e00a17a
            """,
        )

        # Specific options
        self.opt_parser.add_option(
            "-b",
            "--blocking",
            dest="blocking",
            default=False,
            action="store_true",
            help="blocking mode. Wait until the operation completes.",
        )
        self.opt_parser.add_option(
            "-i",
            "--interval",
            dest="poll_interval",
            type="int",
            default=30,
            help="interval between two poll operations in blocking mode.",
        )
        self.opt_parser.add_option(
            "-e",
            "--expire",
            dest="proxy_lifetime",
            type="int",
            default=420,
            help="expiration time of the delegation in minutes.",
        )
        self.opt_parser.add_option(
            "--job-metadata", dest="job_metadata", help="transfer job metadata."
        )
        self.opt_parser.add_option(
            "--file-metadata", dest="file_metadata", help="file metadata."
        )
        self.opt_parser.add_option(
            "-S",
            "--spacetoken",
            dest="spacetoken",
            help="the space token or its description.",
        )
        self.opt_parser.add_option(
            "--dry-run",
            dest="dry_run",
            default=False,
            action="store_true",
            help="do not send anything, just print the JSON message.",
        )
        self.opt_parser.add_option(
            "-f",
            "--file",
            dest="bulk_file",
            type="string",
            help="Name of configuration file",
        )
        self.opt_parser.add_option(
            "--retry",
            dest="retry",
            type="int",
            default=0,
            help="Number of retries. If 0, the server default will be used."
            "If negative, there will be no retries.",
        )
        self.opt_parser.add_option(
            "--cloud-credentials",
            dest="cloud_cred",
            default=None,
            help="use cloud credentials for the job (i.e. dropbox).",
        )

    def validate(self):
        if not self.options.bulk_file:
            if len(self.args) < 1:
                self.logger.critical("Need at least a surl")
                sys.exit(1)

        if self.options.verbose:
            self.logger.setLevel(logging.DEBUG)

    def _build_delete(self):
        if self.options.bulk_file:
            files = list(filter(len, open(self.options.bulk_file).readlines()))
            if len(files):
                return files
            else:
                self.logger.critical("Could not find any file to delete")
                sys.exit(1)
        else:
            return self.args

    def _do_submit(self, context):

        delegator = Delegator(context)
        delegator.delegate(timedelta(minutes=self.options.proxy_lifetime))

        submitter = Submitter(context)

        job_id = submitter.submit(
            delete=self._build_delete(),
            spacetoken=self.options.spacetoken,
            job_metadata=_metadata(self.options.job_metadata),
            retry=self.options.retry,
            credential=self.options.cloud_cred,
        )

        if self.options.json:
            self.logger.info(json.dumps(job_id))
        else:
            self.logger.info("Job successfully submitted.")
            self.logger.info("Job id: %s" % job_id)

        if job_id and self.options.blocking:
            inquirer = Inquirer(context)
            job = inquirer.get_job_status(job_id)
            while job["job_state"] in [
                "SUBMITTED",
                "READY",
                "STAGING",
                "DELETE",
                "ARCHIVING",
                "QOS_TRANSITION",
                "QOS_REQUEST_SUBMITTED",
            ]:
                self.logger.info("Job in state %s" % job["job_state"])
                time.sleep(self.options.poll_interval)
                job = inquirer.get_job_status(job_id)

            self.logger.info("Job finished with state %s" % job["job_state"])
            if job["reason"]:
                self.logger.info("Reason: %s" % job["reason"])

        return job_id

    def _do_dry_run(self, context):

        submitter = Submitter(context)
        print(
            submitter.build_submission(
                delete=self._build_delete(),
                spacetoken=self.options.spacetoken,
                job_metadata=_metadata(self.options.job_metadata),
                retry=self.options.retry,
                credential=self.options.cloud_cred,
            )
        )
        return None

    def run(self):
        context = self._create_context()
        if not self.options.dry_run:
            return self._do_submit(context)
        else:
            return self._do_dry_run(context)



#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import timedelta
import json
import logging
import sys
import time

from .base import Base
from fts3.rest.client import Submitter, Delegator, Inquirer

DEFAULT_PARAMS = {
    "checksum": "ADLER32",
    "overwrite": False,
    "overwrite_on_retry": False,
    "reuse": False,
    "job_metadata": None,
    "file_metadata": None,
    "filesize": None,
    "gridftp": None,
    "spacetoken": None,
    "source_spacetoken": None,
    "verify_checksum": "n",
    "copy_pin_lifetime": -1,
    "bring_online": -1,
    "dst_file_report": False,
    "archive_timeout": -1,
    "timeout": None,
    "fail_nearline": False,
    "retry": 0,
    "multihop": False,
    "credential": None,
    "nostreams": None,
    "s3alternate": False,
    "target_qos": None,
    "ipv4": False,
    "ipv6": False,
    "buffer_size": None,
}


def _metadata(data):
    try:
        return json.loads(data)
    except Exception:
        return str(data)


class JobSubmitter(Base):
    def __init__(self):
        super(JobSubmitter, self).__init__(
            extra_args="SOURCE DESTINATION [CHECKSUM]",
            description="""
            This command can be used to submit new jobs to FTS3. It supports simple and bulk submissions. The bulk
            format is as follows:

            ```json
            {
              "files": [
                {
                  "sources": [
                    "gsiftp://source.host/file"
                  ],
                  "destinations": [
                    "gsiftp://destination.host/file"
                  ],
                  "metadata": "file-metadata",
                  "checksum": "ADLER32:1234",
                  "filesize": 1024
                },
                {
                  "sources": [
                    "gsiftp://source.host/file2"
                  ],
                  "destinations": [
                    "gsiftp://destination.host/file2"
                  ],
                  "metadata": "file2-metadata",
                  "checksum": "ADLER32:4321",
                  "filesize": 2048,
                  "activity": "default"
                }
              ]
            }
            ```
            """,
            example="""
            $ %(prog)s -s https://fts3-devel.cern.ch:8446 gsiftp://source.host/file gsiftp://destination.host/file
            Job successfully submitted.
            Job id: 9fee8c1e-c46d-11e3-8299-02163e00a17a

            $ %(prog)s -s https://fts3-devel.cern.ch:8446 -f bulk.json
            Job successfully submitted.
            Job id: 9fee8c1e-c46d-11e3-8299-02163e00a17a
            """,
        )

        # Specific options
        self.opt_parser.add_option(
            "-b",
            "--blocking",
            dest="blocking",
            default=False,
            action="store_true",
            help="blocking mode. Wait until the operation completes.",
        )
        self.opt_parser.add_option(
            "-i",
            "--interval",
            dest="poll_interval",
            type="int",
            default=30,
            help="interval between two poll operations in blocking mode.",
        )
        self.opt_parser.add_option(
            "-e",
            "--expire",
            dest="proxy_lifetime",
            type="int",
            default=420,
            help="expiration time of the delegation in minutes.",
        )
        self.opt_parser.add_option(
            "--delegate-when-lifetime-lt",
            type=int,
            default=120,
            help="delegate the proxy when the remote lifetime is less than this value (in minutes)",
        )
        self.opt_parser.add_option(
            "-o",
            "--overwrite",
            dest="overwrite",
            action="store_true",
            help="overwrite files.",
        )
        self.opt_parser.add_option(
            "--overwrite--on-retry",
            dest="overwrite_on_retry",
            action="store_true",
            help="overwrite files.",
        )
        self.opt_parser.add_option(
            "-r",
            "--reuse",
            dest="reuse",
            action="store_true",
            help="enable session reuse for the transfer job.",
        )
        self.opt_parser.add_option(
            "--job-metadata", dest="job_metadata", help="transfer job metadata."
        )
        self.opt_parser.add_option(
            "--file-metadata", dest="file_metadata", help="file metadata."
        )
        self.opt_parser.add_option(
            "--file-size", dest="file_size", type="long", help="file size (in Bytes)"
        )
        self.opt_parser.add_option(
            "-g", "--gparam", dest="gridftp_params", help="GridFTP parameters."
        )
        self.opt_parser.add_option(
            "-t",
            "--dest-token",
            dest="destination_token",
            help="the destination space token or its description.",
        )
        self.opt_parser.add_option(
            "-S",
            "--source-token",
            dest="source_token",
            help="the source space token or its description.",
        )
        self.opt_parser.add_option(
            "-K",
            "--compare-checksum",
            dest="compare_checksum",
            default=False,
            action="store_true",
            help="deprecated: compare checksums between source and destination.",
        )
        self.opt_parser.add_option(
            "-C",
            "--checksum-mode",
            dest="checksum_mode",
            type="string",
            help="compare checksums in source, target, both or none.",
        )
        self.opt_parser.add_option(
            "--copy-pin-lifetime",
            dest="pin_lifetime",
            type="long",
            help="pin lifetime of the copy in seconds.",
        )
        self.opt_parser.add_option(
            "--bring-online",
            dest="bring_online",
            type="long",
            help="bring online timeout in seconds.",
        )
        self.opt_parser.add_option(
            "--dst-file-report",
            dest="dst_file_report",
            default=False,
            action="store_true",
            help="report on the destination tape file if it already exists and overwrite is off.",
        )
        self.opt_parser.add_option(
            "--archive-timeout",
            dest="archive_timeout",
            type="long",
            help="archive timeout in seconds.",
        )
        self.opt_parser.add_option(
            "--timeout",
            dest="timeout",
            type="long",
            help="transfer timeout in seconds.",
        )
        self.opt_parser.add_option(
            "--fail-nearline",
            dest="fail_nearline",
            action="store_true",
            help="fail the transfer is the file is nearline.",
        )
        self.opt_parser.add_option(
            "--dry-run",
            dest="dry_run",
            default=False,
            action="store_true",
            help="do not send anything, just print the JSON message.",
        )
        self.opt_parser.add_option(
            "-f",
            "--file",
            dest="bulk_file",
            type="string",
            help="Name of configuration file",
        )
        self.opt_parser.add_option(
            "--retry",
            dest="retry",
            type="int",
            help="Number of retries. If 0, the server default will be used."
            "If negative, there will be no retries.",
        )
        self.opt_parser.add_option(
            "-m",
            "--multi-hop",
            dest="multihop",
            action="store_true",
            help="submit a multihop transfer.",
        )
        self.opt_parser.add_option(
            "--cloud-credentials",
            dest="cloud_cred",
            help="use cloud credentials for the job (i.e. dropbox).",
        )
        self.opt_parser.add_option(
            "--nostreams", dest="nostreams", help="number of streams"
        )
        self.opt_parser.add_option(
            "--ipv4", dest="ipv4", action="store_true", help="force ipv4"
        )
        self.opt_parser.add_option(
            "--ipv6", dest="ipv6", action="store_true", help="force ipv6"
        )
        self.opt_parser.add_option(
            "--s3alternate",
            dest="s3alternate",
            action="store_true",
            help="use S3 alternate URL",
        )
        self.opt_parser.add_option(
            "--target-qos",
            dest="target_qos",
            type="string",
            help="define the target QoS for this transfer for CDMI endpoints",
        )
        self.opt_parser.add_option(
            "--buffer-size",
            "--buff-size",
            dest="buffer_size",
            type=int,
            help="TCP buffer size (expressed in bytes) that will be used for the given transfer job",
        )

    def validate(self):
        self.checksum = None
        if not self.options.bulk_file:
            if len(self.args) < 2:
                self.logger.critical("Need a source and a destination")
                sys.exit(1)
            elif len(self.args) == 2:
                (self.source, self.destination) = self.args
            elif len(self.args) == 3:
                (self.source, self.destination, self.checksum) = self.args
            else:
                self.logger.critical("Too many parameters")
                sys.exit(1)

        self._prepare_options()
        if self.params["ipv4"] and self.params["ipv6"]:
            self.opt_parser.error("ipv4 and ipv6 can not be used at the same time")
        if self.params["overwrite"] and self.params["overwrite_on_retry"]:
            self.opt_parser.error(
                "overwrite and overwrite-on-retry can not be used at the same time"
            )

    def _build_transfers(self):
        if self.options.bulk_file:
            with open(self.options.bulk_file, "r") as file:
                filecontent = file.read()
                bulk = json.loads(filecontent)
            if "files" in bulk:
                return bulk["files"]
            elif "Files" in bulk:
                return bulk["Files"]
            else:
                self.logger.critical("Could not find any transfers")
                sys.exit(1)
        else:
            return [{"sources": [self.source], "destinations": [self.destination]}]

    def _build_params(self, **kwargs):
        params = dict()
        params.update(DEFAULT_PARAMS)

        if self.options.bulk_file:
            with open(self.options.bulk_file, "r") as file:
                filecontent = file.read()
                bulk = json.loads(filecontent)
            if "params" in bulk:
                params.update(bulk["params"])

        # Apply command-line parameters
        for k, v in iter(kwargs.items()):
            if v is not None:
                params[k] = v

        # JSONify metadata
        if params["job_metadata"] is not None:
            params["job_metadata"] = _metadata(params["job_metadata"])
        if params["file_metadata"] is not None:
            params["file_metadata"] = _metadata(params["file_metadata"])
        return params

    def _prepare_options(self):
        # Backwards compatibility: compare_checksum parameter
        # Note: compare_checksum has higher priority than checksum_mode
        if self.options.compare_checksum:
            checksum_mode = "both"
        elif self.options.checksum_mode:
            checksum_mode = self.options.checksum_mode
        else:
            checksum_mode = "none"

        self.transfers = self._build_transfers()
        self.params = self._build_params(
            checksum=self.checksum,
            bring_online=self.options.bring_online,
            dst_file_report=self.options.dst_file_report,
            archive_timeout=self.options.archive_timeout,
            timeout=self.options.timeout,
            verify_checksum=checksum_mode[0],
            spacetoken=self.options.destination_token,
            source_spacetoken=self.options.source_token,
            fail_nearline=self.options.fail_nearline,
            file_metadata=_metadata(self.options.file_metadata),
            filesize=self.options.file_size,
            gridftp=self.options.gridftp_params,
            job_metadata=_metadata(self.options.job_metadata),
            overwrite=self.options.overwrite,
            overwrite_on_retry=self.options.overwrite_on_retry,
            copy_pin_lifetime=self.options.pin_lifetime,
            reuse=self.options.reuse,
            retry=self.options.retry,
            multihop=self.options.multihop,
            credential=self.options.cloud_cred,
            nostreams=self.options.nostreams,
            ipv4=self.options.ipv4,
            ipv6=self.options.ipv6,
            s3alternate=self.options.s3alternate,
            target_qos=self.options.target_qos,
            buffer_size=self.options.buffer_size,
        )

    def _do_submit(self, context):
        if not self.options.access_token:
            delegator = Delegator(context)
            delegator.delegate(
                timedelta(minutes=self.options.proxy_lifetime),
                delegate_when_lifetime_lt=timedelta(
                    minutes=self.options.delegate_when_lifetime_lt
                ),
            )
            self.logger.debug(
                "Delegation termination time: %s"
                % delegator.get_info()["termination_time"]
            )

        submitter = Submitter(context)
        job_id = submitter.submit(transfers=self.transfers, params=self.params)

        if self.options.json:
            self.logger.info(json.dumps(job_id))
        else:
            self.logger.info("Job successfully submitted.")
            self.logger.info("Job id: %s" % job_id)
        if job_id and self.options.blocking:
            inquirer = Inquirer(context)
            job = inquirer.get_job_status(job_id)
            while job["job_state"] in [
                "SUBMITTED",
                "READY",
                "STAGING",
                "ACTIVE",
                "ARCHIVING",
                "QOS_TRANSITION",
                "QOS_REQUEST_SUBMITTED",
            ]:
                self.logger.info("Job in state %s" % job["job_state"])
                time.sleep(self.options.poll_interval)
                job = inquirer.get_job_status(job_id)

            self.logger.info("Job finished with state %s" % job["job_state"])
            if job["reason"]:
                self.logger.info("Reason: %s" % job["reason"])

        return job_id

    def _do_dry_run(self, context):
        submitter = Submitter(context)
        print(submitter.build_submission(transfers=self.transfers, params=self.params))
        return None

    def run(self):
        context = self._create_context()
        if not self.options.dry_run:
            return self._do_submit(context)
        else:
            return self._do_dry_run(context)



#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from .banning import Banning
from .delegator import Delegator
from .jobcanceller import JobCanceller
from .joblister import JobLister
from .jobshower import JobShower
from .jobsubmitter import JobSubmitter
from .jobdeletionsubmitter import JobDeletionSubmitter
from .serverstatus import ServerStatus
from .whoami import WhoAmI
import logging
import sys


class FTS3CliFormatter(logging.Formatter):
    def format(self, record):

        if record.levelno == logging.CRITICAL:
            self._fmt = "Error: %(msg)s"
        elif record.levelno == logging.WARNING:
            self._fmt = "# Warning: %(msg)s"
        elif record.levelno == logging.DEBUG:
            self._fmt = "# %(msg)s"
        else:
            self._fmt = "%(msg)s"

        return logging.Formatter.format(self, record)


class FTS3CliFilter(object):
    def __init__(self, includes):
        self.includes = includes

    def filter(self, record):
        return record.levelno in self.includes


class FTS3CliFilterExclude(object):
    def __init__(self, excludes):
        self.excludes = excludes

    def filter(self, record):
        return record.levelno not in self.excludes


fmt = FTS3CliFormatter()
stdout_handler = logging.StreamHandler(sys.stdout)
stderr_handler = logging.StreamHandler(sys.stderr)

stdout_handler.setFormatter(fmt)
stdout_handler.addFilter(FTS3CliFilterExclude([logging.WARNING, logging.DEBUG]))
stderr_handler.setFormatter(fmt)
stderr_handler.addFilter(FTS3CliFilter([logging.WARNING, logging.DEBUG]))

logging.root.addHandler(stdout_handler)
logging.root.addHandler(stderr_handler)

logging.root.setLevel(logging.INFO)



#   Copyright 2014-2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import sys

from .base import Base
from fts3.rest.client import Ban


class Banning(Base):
    def __init__(self):
        super(Banning, self).__init__(
            description="Ban and unban storage elements and users",
            example="""
                $ %(prog)s -s https://fts3-devel.cern.ch:8446 --storage gsiftp://sample
                No jobs affected
                $ %(prog)s -s https://fts3-devel.cern.ch:8446 --storage gsiftp://sample --unban
                $
            """,
        )

        self.opt_parser.add_option("--storage", dest="storage", help="storage element")
        self.opt_parser.add_option("--user", dest="user_dn", help="user dn")
        self.opt_parser.add_option(
            "--unban",
            dest="unban",
            default=False,
            action="store_true",
            help="unban instead of ban",
        )

        self.opt_parser.add_option(
            "--status",
            dest="status",
            default="cancel",
            help="status of the jobs that are already in the queue: cancel or wait",
        )
        self.opt_parser.add_option(
            "--timeout",
            dest="timeout",
            default=0,
            help="the timeout for the jobs that are already in the queue if status is wait",
        )
        self.opt_parser.add_option(
            "--allow-submit",
            dest="allow_submit",
            default=False,
            action="store_true",
            help="allow submissions if status is wait",
        )

    def validate(self):
        # Some sanity checks
        # This are checked server side anyway (or so they should) but we can shorcurt here
        self.options.status = self.options.status.lower()
        if self.options.status not in ["cancel", "wait"]:
            self.logger.critical("Status can only be cancel or wait")
            sys.exit(1)

        if self.options.status == "cancel":
            if self.options.allow_submit:
                self.logger.critical(
                    "--allow-submit can only be used with --status=wait"
                )
                sys.exit(1)
        else:
            if self.options.user_dn:
                self.logger.critical("--user only accept cancel")
                sys.exit(1)

        if (not self.options.storage and not self.options.user_dn) or (
            self.options.storage and self.options.user_dn
        ):
            self.logger.critical("Need to specify only one of --storage or --user")
            sys.exit(1)

    def run(self):
        context = self._create_context()
        ban = Ban(context)

        affected_jobs = None
        if self.options.storage:
            if self.options.unban:
                ban.unban_se(self.options.storage)
            else:
                affected_jobs = ban.ban_se(
                    self.options.storage,
                    self.options.status,
                    self.options.timeout,
                    self.options.allow_submit,
                )
        else:
            if self.options.unban:
                ban.unban_dn(self.options.user_dn)
            else:
                affected_jobs = ban.ban_dn(self.options.user_dn)

        if affected_jobs:
            self.logger.info("Affected jobs:")
            for j in affected_jobs:
                self.logger.info(j)
        elif not self.options.unban:
            self.logger.info("No jobs affected")



#   Copyright 2015-2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
from .base import Base


class ServerStatus(Base):
    def __init__(self):
        super(ServerStatus, self).__init__(
            description="Use this command to check on the service status.",
        )

        self.opt_parser.add_option(
            "-H",
            "--host",
            dest="host",
            default=None,
            help="limit the output to a given host",
        )
        self.opt_parser.add_option(
            "--is-active",
            dest="is_active",
            default=False,
            action="store_true",
            help="the tool will return < 0 on error, 0 if nothing is active, "
            "1 if there are active transfers, 2 if there are active staging, 3 if there are both ",
        )

    def run(self):
        context = self._create_context()
        host_activity = json.loads(context.get("/status/hosts"))
        hosts = [self.options.host] if self.options.host else host_activity.keys()
        total_count = dict(active=0, staging=0)
        for host in hosts:
            self.logger.info(host)
            for state, count in host_activity.get(host, {}).items():
                self.logger.info("\t%s: %d" % (state, count))
                total_count[state] += count

        if self.options.is_active:
            return ((total_count["active"] > 0) * 1) + (
                (total_count["staging"] > 0) * 2
            )
        else:
            return 0



#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import sys

from fts3.rest.client import Inquirer
from .base import Base
from .utils import *


class JobShower(Base):
    def __init__(self):
        super(JobShower, self).__init__(
            extra_args="JOB_ID",
            description="This command can be used to check the current status of a given job",
            example="""
            $ %(prog)s -s https://fts3-devel.cern.ch:8446 c079a636-c363-11e3-b7e5-02163e009f5a
            Request ID: c079a636-c363-11e3-b7e5-02163e009f5a
            Status: FINISHED
            Client DN: /DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=saketag/CN=678984/CN=Alejandro Alvarez Ayllon
            Reason:
            Submission time: 2014-04-13T23:31:34
            Priority: 3
            VO Name: dteam
            """,
        )

    def validate(self):
        if len(self.args) == 0:
            self.logger.critical("Need a job id")
            sys.exit(1)

    def run(self):
        job_id = self.args[0]
        context = self._create_context()

        inquirer = Inquirer(context)
        job = inquirer.get_job_status(job_id, list_files=self.options.json)

        if not self.options.json:
            self.logger.info(job_human_readable(job))
        else:
            self.logger.info(job_as_json(job))



#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import datetime
import getpass
import json
import logging
import os
import sys
from urllib.parse import quote

from M2Crypto import X509, RSA, EVP, BIO
from M2Crypto.ASN1 import UTC
from fts3 import __version__ as CLIENT_VERSION

from .exceptions import *
from .request import Request

log = logging.getLogger(__name__)


# Return a list of certificates from the file
def _get_x509_list(cert):
    x509_list = []
    fd = BIO.openfile(cert, "rb")
    cert = X509.load_cert_bio(fd)
    try:
        while True:
            x509_list.append(cert)
            log.debug("Loaded " + cert.get_subject().as_text())
            cert = X509.load_cert_bio(fd)
    except X509.X509Error:
        # When there are no more certs, this is what we get, so it is fine
        pass
    except BIO.BIOError:
        # When there are no more certs, this is what we get, so it is fine
        # Python 2.4
        pass

    del fd
    return x509_list


def _get_default_proxy():
    """
    Returns the default proxy location
    """
    return "/tmp/x509up_u%d" % os.geteuid()


class Context(object):
    def _read_passwd_from_stdin(self, *args, **kwargs):
        if not self.passwd:
            self.passwd = getpass.getpass("Private key password: ")
        return self.passwd

    def _set_x509(self, ucert, ukey):
        default_proxy_location = _get_default_proxy()

        # User certificate and key locations
        if ucert and not ukey:
            ukey = ucert
        elif not ucert:
            if "X509_USER_PROXY" in os.environ:
                ukey = ucert = os.environ["X509_USER_PROXY"]
            elif os.path.exists(default_proxy_location):
                ukey = ucert = default_proxy_location
            elif "X509_USER_CERT" in os.environ:
                ucert = os.environ["X509_USER_CERT"]
                ukey = os.environ.get("X509_USER_KEY", ucert)
            elif os.path.exists("/etc/grid-security/hostcert.pem") and os.path.exists(
                "/etc/grid-security/hostkey.pem"
            ):
                ucert = "/etc/grid-security/hostcert.pem"
                ukey = "/etc/grid-security/hostkey.pem"

        if ucert and ukey:
            self.x509_list = _get_x509_list(ucert)
            self.x509 = self.x509_list[0]
            not_after = self.x509.get_not_after()
            try:
                not_after = not_after.get_datetime()
            except Exception:
                # Ugly hack for Python 2.4
                import time

                not_after = datetime.fromtimestamp(
                    time.mktime(time.strptime(str(not_after), "%b %d %H:%M:%S %Y %Z")),
                    tz=UTC,
                )

            if not_after < datetime.now(UTC):
                raise Exception("Proxy expired!")

            try:
                self.rsa_key = RSA.load_key(ukey, self._read_passwd_from_stdin)
            except RSA.RSAError as e:
                raise RSA.RSAError("Could not load %s: %s" % (ukey, str(e)))
            except Exception as e:
                raise Exception("Could not load %s: %s" % (ukey, str(e)))

            self.evp_key = EVP.PKey()
            self.evp_key.assign_rsa(self.rsa_key)

            self.ucert = ucert
            self.ukey = ukey
        else:
            self.ucert = self.ukey = None

        if not self.ucert and not self.ukey:
            log.warning("No user certificate given!")
        else:
            log.debug("User certificate: %s" % self.ucert)
            log.debug("User private key: %s" % self.ukey)

    def _set_endpoint(self, endpoint):
        self.endpoint = endpoint
        if self.endpoint.endswith("/"):
            self.endpoint = self.endpoint[:-1]

    def _validate_endpoint(self):
        try:
            endpoint_info = json.loads(self.get("/"))
            endpoint_info["url"] = self.endpoint
        except FTS3ClientException:
            raise
        except Exception as e:
            raise BadEndpoint("%s (%s)" % (self.endpoint, str(e))).with_traceback(
                sys.exc_info()[2]
            )
        return endpoint_info

    def _set_user_agent(self, user_agent=None):
        if user_agent is None:
            self.user_agent = "fts-python-bindings/" + CLIENT_VERSION
        else:
            self.user_agent = user_agent

    def __init__(
        self,
        endpoint,
        ucert=None,
        ukey=None,
        verify=True,
        access_token=None,
        no_creds=False,
        capath=None,
        request_class=Request,
        connectTimeout=30,
        timeout=30,
        user_agent=None,
    ):
        self.passwd = None
        self.access_method = None

        self._set_user_agent(user_agent)
        self._set_endpoint(endpoint)
        if no_creds:
            self.ucert = self.ukey = self.access_token = None
        else:
            self.access_token = access_token
            if self.access_token:
                self.ucert = None
                self.ukey = None
                self.access_method = "oauth2"
            else:
                self._set_x509(ucert, ukey)
                self.access_method = "X509"

        self._requester = request_class(
            self.ucert,
            self.ukey,
            passwd=self.passwd,
            verify=verify,
            access_token=self.access_token,
            capath=capath,
            connectTimeout=connectTimeout,
            timeout=timeout,
        )

        self.endpoint_info = self._validate_endpoint()
        # Log obtained information
        log.debug("Using endpoint: %s" % self.endpoint_info["url"])
        log.debug(
            "REST API version: %(major)d.%(minor)d.%(patch)d"
            % self.endpoint_info["api"]
        )

    def get_endpoint_info(self):
        return self.endpoint_info

    def get(self, path, args=None):
        if args:
            query = "&".join("%s=%s" % (k, quote(v)) for k, v in args.items())
            path += "?" + query
        return self._requester.method(
            "GET",
            "%s/%s" % (self.endpoint, path),
            headers={"User-Agent": self.user_agent},
        )

    def put(self, path, body):
        return self._requester.method(
            "PUT",
            "%s/%s" % (self.endpoint, path),
            body=body,
            headers={"User-Agent": self.user_agent},
        )

    def delete(self, path):
        return self._requester.method(
            "DELETE",
            "%s/%s" % (self.endpoint, path),
            headers={"User-Agent": self.user_agent},
        )

    def post_json(self, path, body):
        if not isinstance(body, str):
            body = json.dumps(body)
        headers = {"Content-Type": "application/json", "User-Agent": self.user_agent}
        return self._requester.method(
            "POST",
            "%s/%s" % (self.endpoint, path),
            body=body,
            headers=headers,
        )



# pylint: skip-file
# Pylint will detect missing M2Crypto members

#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import ctypes
import json
import logging
import sys
import time

from datetime import datetime, timedelta
from M2Crypto import X509, ASN1, m2, Err
from M2Crypto.ASN1 import UTC

# I deleted from pytz import utc as UTC, as I don't see any reason to keep it

from .exceptions import *

log = logging.getLogger(__name__)

# See https://bugzilla.osafoundation.org/show_bug.cgi?id=7530
# for an explanation on all this mess
# TL;DR: M2Crypto fails to properly initialize the internal structure, which
# results eventually in segfaults
class M2Ctx(ctypes.Structure):
    _fields_ = [
        ("flags", ctypes.c_int),
        ("issuer_cert", ctypes.c_void_p),
        ("subject_cert", ctypes.c_void_p),
        ("subject_req", ctypes.c_void_p),
        ("crl", ctypes.c_void_p),
        ("db_meth", ctypes.c_void_p),
        ("db", ctypes.c_void_p),
    ]


def _init_m2_ctx(m2_ctx, issuer=None):
    ctx = M2Ctx.from_address(int(m2_ctx))
    ctx.flags = 0
    ctx.subject_cert = None
    ctx.subject_req = None
    ctx.crl = None
    if issuer is None:
        ctx.issuer_cert = None
    else:
        ctx.issuer_cert = int(issuer.x509)


def _workaround_new_extension(name, value, critical=False, issuer=None, _pyfree=1):
    # m2crypto removes x509v3_lhash with 0.25.1
    try:
        ctx = m2.x509v3_set_nconf()
        if ctx is None:
            raise MemoryError()
        _init_m2_ctx(ctx, issuer)
        x509_ext_ptr = m2.x509v3_ext_conf(None, ctx, name, value)
    except AttributeError:
        lhash = m2.x509v3_lhash()
        ctx = m2.x509v3_set_conf_lhash(lhash)
        _init_m2_ctx(ctx, issuer)
        x509_ext_ptr = m2.x509v3_ext_conf(lhash, ctx, name, value)

    if x509_ext_ptr is None:
        raise Exception("Could not create the X509v3 extension")

    x509_ext = X509.X509_Extension(x509_ext_ptr, _pyfree)
    x509_ext.set_critical(critical)
    return x509_ext


def _m2crypto_extensions_broken():
    # SL5 is out support
    return False


def _add_rfc3820_extensions(proxy):
    proxy.add_ext(
        X509.new_extension(
            "proxyCertInfo", "critical,language:Inherit all", critical=True
        )
    )


class Delegator(object):

    nid = {
        "C": m2.NID_countryName,
        "SP": m2.NID_stateOrProvinceName,
        "ST": m2.NID_stateOrProvinceName,
        "stateOrProvinceName": m2.NID_stateOrProvinceName,
        "L": m2.NID_localityName,
        "localityName": m2.NID_localityName,
        "O": m2.NID_organizationName,
        "organizationName": m2.NID_organizationName,
        "OU": m2.NID_organizationalUnitName,
        "organizationUnitName": m2.NID_organizationalUnitName,
        "CN": m2.NID_commonName,
        "commonName": m2.NID_commonName,
        "Email": m2.NID_pkcs9_emailAddress,
        "emailAddress": m2.NID_pkcs9_emailAddress,
        "serialNumber": m2.NID_serialNumber,
        "SN": m2.NID_surname,
        "surname": m2.NID_surname,
        "GN": m2.NID_givenName,
        "givenName": m2.NID_givenName,
        "DC": 391,
    }

    def __init__(self, context):
        self.context = context

    def _get_delegation_id(self):
        r = json.loads(self.context.get("/whoami"))
        return r["delegation_id"]

    def _get_remaining_life(self, delegation_id):
        r = self.get_info(delegation_id)
        if not r:
            return None
        else:
            expiration_time = datetime.strptime(
                r["termination_time"], "%Y-%m-%dT%H:%M:%S"
            )
            return expiration_time - datetime.utcnow()

    def _get_proxy_request(self, delegation_id):
        request_url = "/delegation/" + delegation_id + "/request"
        request_pem = self.context.get(request_url)
        x509_request = X509.load_request_string(request_pem)
        if x509_request.verify(x509_request.get_pubkey()) != 1:
            raise ServerError(
                "Error verifying signature on the request:", Err.get_error()
            )
        # Return
        return x509_request

    def _sign_request(self, x509_request, lifetime):
        not_before = ASN1.ASN1_UTCTIME()
        not_before.set_datetime(datetime.now(UTC))
        not_after = ASN1.ASN1_UTCTIME()
        not_after.set_datetime(datetime.now(UTC) + lifetime)

        proxy_subject = X509.X509_Name()
        for entry in self.context.x509.get_subject():
            ret = m2.x509_name_add_entry(proxy_subject._ptr(), entry._ptr(), -1, 0)
            if ret == 0:
                raise Exception(
                    "%s: '%s'" % (m2.err_reason_error_string(m2.err_get_error()), entry)
                )

        proxy = X509.X509()
        proxy.set_serial_number(self.context.x509.get_serial_number())
        proxy.set_version(x509_request.get_version())
        proxy.set_issuer(self.context.x509.get_subject())
        proxy.set_pubkey(x509_request.get_pubkey())

        # Extensions are broken in SL5!!
        if _m2crypto_extensions_broken():
            log.warning("X509v3 extensions disabled!")
        else:
            # X509v3 Basic Constraints
            proxy.add_ext(
                X509.new_extension("basicConstraints", "CA:FALSE", critical=True)
            )
            # X509v3 Key Usage
            proxy.add_ext(
                X509.new_extension(
                    "keyUsage", "Digital Signature, Key Encipherment", critical=True
                )
            )
            # X509v3 Authority Key Identifier
            identifier_ext = _workaround_new_extension(
                "authorityKeyIdentifier",
                "keyid",
                critical=False,
                issuer=self.context.x509,
            )
            proxy.add_ext(identifier_ext)

        any_rfc_proxies = False
        # FTS-1217 Ignore the user input and select the min proxy lifetime available on the list
        min_cert_lifetime = self.context.x509_list[0].get_not_after()
        for cert in self.context.x509_list:
            if cert.get_not_after().get_datetime() < min_cert_lifetime.get_datetime():
                not_after = cert.get_not_after()
                min_cert_lifetime = cert.get_not_after()
            try:
                cert.get_ext("proxyCertInfo")
                any_rfc_proxies = True
            except Exception:
                pass

        proxy.set_not_after(not_after)
        proxy.set_not_before(not_before)

        if any_rfc_proxies:
            if _m2crypto_extensions_broken():
                raise NotImplementedError(
                    "X509v3 extensions are disabled, so RFC proxies can not be generated!"
                )
            else:
                _add_rfc3820_extensions(proxy)

        if any_rfc_proxies:
            m2.x509_name_set_by_nid(
                proxy_subject._ptr(),
                X509.X509_Name.nid["commonName"],
                str(int(time.time())).encode("utf-8"),
            )
        else:
            m2.x509_name_set_by_nid(
                proxy_subject._ptr(), X509.X509_Name.nid["commonName"], b"proxy"
            )

        proxy.set_subject(proxy_subject)
        proxy.set_version(2)
        proxy.sign(self.context.evp_key, "sha1")

        return proxy

    def _put_proxy(self, delegation_id, x509_proxy):
        self.context.put("/delegation/" + delegation_id + "/credential", x509_proxy)

    def _full_proxy_chain(self, x509_proxy):
        chain = x509_proxy.as_pem()
        for cert in self.context.x509_list:
            chain += cert.as_pem()
        return chain

    def get_info(self, delegation_id=None):
        if delegation_id is None:
            delegation_id = self._get_delegation_id()
        return json.loads(self.context.get("/delegation/" + delegation_id))

    def delegate(
        self,
        lifetime=timedelta(hours=7),
        force=False,
        delegate_when_lifetime_lt=timedelta(hours=2),
    ):
        try:
            delegation_id = self._get_delegation_id()
            log.debug("Delegation ID: " + delegation_id)

            remaining_life = self._get_remaining_life(delegation_id)

            if remaining_life is None:
                log.debug("No previous delegation found")
            elif remaining_life <= timedelta(0):
                log.debug("The delegated credentials expired")
            elif (
                self.context.access_method == "oauth2"
                or remaining_life >= delegate_when_lifetime_lt
            ):
                if self.context.access_method == "oauth2" or not force:
                    log.debug("Not bothering doing the delegation")
                    return delegation_id
                else:
                    log.debug("Delegation not expired, but this is a forced delegation")

            # Ask for the request
            log.debug("Delegating")
            x509_request = self._get_proxy_request(delegation_id)

            # Sign request
            log.debug("Signing request")
            x509_proxy = self._sign_request(x509_request, lifetime)
            x509_proxy_pem = self._full_proxy_chain(x509_proxy)

            # Send the signed proxy
            self._put_proxy(delegation_id, x509_proxy_pem)

            return delegation_id

        except Exception as e:
            raise ClientError(str(e)).with_traceback(sys.exc_info()[2])



#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from fts3.rest.client import Inquirer
from .base import Base
from .utils import *


class JobLister(Base):
    def __init__(self):
        super(JobLister, self).__init__(
            description="This command can be used to list the running jobs, allowing to filter by user dn or vo name",
            example="""
            $ %(prog)s -s https://fts3-devel.cern.ch:8446 -o atlas
            Request ID: ff294db7-655a-4c0a-9efb-44a994677bb3
            Status: ACTIVE
            Client DN: /DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=ddmadmin/CN=531497/CN=Robot: ATLAS Data Management
            Reason: None
            Submission time: 2014-04-15T07:05:38
            Priority: 3
            VO Name: atlas

            Request ID: a2e4586c-760a-469e-8303-d0f3d5aadc73
            Status: READY
            Client DN: /DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=ddmadmin/CN=531497/CN=Robot: ATLAS Data Management
            Reason: None
            Submission time: 2014-04-15T07:07:33
            Priority: 3
            VO Name: atlas
            """,
        )
        # Specific options
        self.opt_parser.add_option(
            "-u", "--userdn", dest="user_dn", help="query only for the given user"
        )
        self.opt_parser.add_option(
            "-o", "--voname", dest="vo_name", help="query only for the given VO"
        )
        self.opt_parser.add_option(
            "--source",
            dest="source_se",
            help="query only for the given source storage element",
        )
        self.opt_parser.add_option(
            "--destination",
            dest="dest_se",
            help="query only for the given destination storage element",
        )

    def run(self):
        context = self._create_context()
        inquirer = Inquirer(context)
        job_list = inquirer.get_job_list(
            self.options.user_dn,
            self.options.vo_name,
            self.options.source_se,
            self.options.dest_se,
        )
        if not self.options.json:
            self.logger.info(job_list_human_readable(job_list))
        else:
            self.logger.info(job_list_as_json(job_list))



#   Copyright 2014-2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import timedelta

from fts3.rest.client import Submitter
from fts3.rest.client import ClientError
from .delegate import delegate


class JobIdGenerator:
    standard = "standard"  # Default algorithm using uuid1
    deterministic = "deterministic"  # Deterministic algorithm using uuid5 with base_id+vo+sid given by the user


def cancel(context, job_id, file_ids=None):
    """
    Cancels a job

    Args:
        context: fts3.rest.client.context.Context instance
        job_id:  The job to cancel

    Returns:
        The terminal state in which the job has been left.
        Note that it may not be CANCELED if the job finished already!
    """
    submitter = Submitter(context)
    return submitter.cancel(job_id, file_ids)


def cancel_all(context, vo_name=None):
    """
    Cancel all jobs within a given VO or FTS3 (needs enough privileges)

    Args:
        context: fts3.rest.client.context.Context instance
        vo_name: The VO name, or None to cancell all jobs

    Returns:
        None
    """
    submitter = Submitter(context)
    return submitter.cancel_all(vo_name)


def new_transfer(
    source,
    destination,
    checksum="ADLER32",
    filesize=None,
    metadata=None,
    activity=None,
    selection_strategy="auto",
):
    """
    Creates a new transfer pair

    Args:
        source:             Source SURL
        destination:        Destination SURL
        checksum:           Checksum
        filesize:           File size
        metadata:           Metadata to bind to the transfer
        selection_strategy: selection Strategy to implement for multiple replica Jobs

    Returns:
        An initialized transfer
    """
    transfer = dict(
        sources=[source],
        destinations=[destination],
    )
    if checksum:
        transfer["checksum"] = checksum
    if filesize:
        transfer["filesize"] = filesize
    if metadata:
        transfer["metadata"] = metadata
    if activity:
        transfer["activity"] = activity
    if selection_strategy:
        transfer["selection_strategy"] = selection_strategy

    return transfer


def add_alternative_source(transfer, alt_source):
    """
    Adds an alternative source to a transfer

    Args:
        transfer:   A dictionary created with new_transfer
        alt_source: Alternative source

    Returns:
        For convenience, transfer
    """
    transfer["sources"].push_back(alt_source)
    return transfer


def new_job(
    transfers=None,
    deletion=None,
    verify_checksum=False,
    reuse=None,
    overwrite=False,
    overwrite_on_retry=False,
    multihop=False,
    source_spacetoken=None,
    spacetoken=None,
    bring_online=None,
    dst_file_report=False,
    archive_timeout=None,
    copy_pin_lifetime=None,
    retry=-1,
    retry_delay=0,
    metadata=None,
    priority=None,
    strict_copy=False,
    max_time_in_queue=None,
    timeout=None,
    id_generator=JobIdGenerator.standard,
    sid=None,
    s3alternate=False,
    nostreams=1,
    buffer_size=None,
):
    """
    Creates a new dictionary representing a job

    Args:
        transfers:         Initial list of transfers
        deletion:          Delete files
        verify_checksum:   Enable checksum verification: source, destination, both or none
        reuse:             Enable reuse (all transfers are handled by the same process)
        overwrite:         Overwrite the destinations if exist
        overwrite_on_retry: Enable overwrite files only during FTS retries
        multihop:          Treat the transfer as a multihop transfer
        source_spacetoken: Source space token
        spacetoken:        Destination space token
        bring_online:      Bring online timeout
        dst_file_report:   Report on the destination tape file if it already exists and overwrite is off
        archive_timeout:   Archive timeout
        copy_pin_lifetime: Pin lifetime
        retry:             Number of retries: <0 is no retries, 0 is server default, >0 is whatever value is passed
        metadata:          Metadata to bind to the job
        priority:          Job priority
        max_time_in_queue: Maximum number
        id_generator:      Job id generator algorithm
        sid:               Specific id given by the client
        s3alternate:       Use S3 alternate url schema
        nostreams:         Number of streams
        buffer_size:       Tcp buffer size (in bytes) that will be used for the given transfer-job

    Returns:
        An initialized dictionary representing a job
    """
    if transfers is None and deletion is None:
        raise ClientError("Bad request: No transfers or deletion jobs are provided")
    if transfers is None:
        transfers = []

    if isinstance(verify_checksum, str):
        if verify_checksum not in ("source", "target", "both", "none"):
            raise ClientError(
                "Bad request: verify_checksum does not contain a valid value"
            )

    if overwrite != False and overwrite_on_retry != False:
        raise ClientError(
            "Bad request: overwrite and overwrite-on-retry can not be used at the same time"
        )

    params = dict(
        verify_checksum=verify_checksum,
        reuse=reuse,
        spacetoken=spacetoken,
        bring_online=bring_online,
        dst_file_report=dst_file_report,
        archive_timeout=archive_timeout,
        copy_pin_lifetime=copy_pin_lifetime,
        job_metadata=metadata,
        source_spacetoken=source_spacetoken,
        overwrite=overwrite,
        overwrite_on_retry=overwrite_on_retry,
        multihop=multihop,
        retry=retry,
        retry_delay=retry_delay,
        priority=priority,
        strict_copy=strict_copy,
        max_time_in_queue=max_time_in_queue,
        timeout=timeout,
        id_generator=id_generator,
        sid=sid,
        s3alternate=s3alternate,
        nostreams=nostreams,
        buffer_size=buffer_size,
    )
    job = dict(files=transfers, delete=deletion, params=params)
    return job


def new_staging_job(
    files,
    bring_online=None,
    copy_pin_lifetime=None,
    source_spacetoken=None,
    spacetoken=None,
    metadata=None,
    priority=None,
    id_generator=JobIdGenerator.standard,
    sid=None,
):
    """
        Creates a new dictionary representing a staging job

    Args:
        files:  Array of surls to stage. Each item can be either a string or a dictionary with keys surl and metadata
        bring_online:      Bring online timeout
        copy_pin_lifetime: Pin lifetime
        source_spacetoken: Source space token
        spacetoken: Deletion spacetoken
        metadata:   Metadata to bind to the job
        priority:          Job priority
        id_generator:      Job id generator algorithm
        sid:               Specific id given by the client

    Returns:
        An initialized dictionary representing a staging job
    """
    if bring_online <= 0 and copy_pin_lifetime <= 0:
        raise ClientError(
            "Bad request: bring_online and copy_pin_lifetime are not positive numbers"
        )

    transfers = []
    for trans in files:
        if isinstance(trans, dict):
            surl = trans["surl"]
            meta = trans["metadata"]
        elif isinstance(trans, str):
            surl = trans
            meta = None
        else:
            raise AttributeError("Unexpected input type %s" % type(files))

        transfers.append(new_transfer(source=surl, destination=surl, metadata=meta))

    params = dict(
        source_spacetoken=source_spacetoken,
        spacetoken=spacetoken,
        bring_online=bring_online,
        copy_pin_lifetime=copy_pin_lifetime,
        job_metadata=metadata,
        priority=priority,
        id_generator=id_generator,
        sid=sid,
    )
    job = dict(files=transfers, params=params)
    return job


def new_delete_job(
    files,
    spacetoken=None,
    metadata=None,
    priority=None,
    id_generator=JobIdGenerator.standard,
    sid=None,
):
    """
    Creates a new dictionary representing a deletion job

    Args:
        files:      Array of surls to delete. Each item can be either a string or a dictionary with keys surl and metadata
        spacetoken: Deletion spacetoken
        metadata:   Metadata to bind to the job
        id_generator:    Job id generator algorithm
        sid:    Specific id given by the client

    Return
        An initialized dictionary representing a deletion job
    """
    params = dict(
        source_spacetoken=spacetoken,
        job_metadata=metadata,
        priority=priority,
        id_generator=id_generator,
        sid=sid,
    )
    job = dict(delete=files, params=params)
    return job


def submit(
    context,
    job,
    delegation_lifetime=timedelta(hours=7),
    force_delegation=False,
    delegate_when_lifetime_lt=timedelta(hours=2),
):
    """
    Submits a job

    Args:
        context: fts3.rest.client.context.Context instance
        job:     Dictionary representing the job
        delegation_lifetime: Delegation lifetime
        force_delegation:    Force delegation even if there is a valid proxy
        delegate_when_lifetime_lt: If the remaining lifetime on the delegated proxy is less than this interval,
                  do a new delegation

    Returns:
        The job id
    """
    delegate(context, delegation_lifetime, force_delegation, delegate_when_lifetime_lt)
    submitter = Submitter(context)
    params = job.get("params", {})
    return submitter.submit(
        transfers=job.get("files", None),
        delete=job.get("delete", None),
        staging=job.get("staging", None),
        **params,
    )



#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
import os
import urllib3
import requests

from .exceptions import *


class Request:
    def __init__(
        self,
        ucert,
        ukey,
        capath=None,
        passwd=None,
        verify=True,
        access_token=None,
        connectTimeout=30,
        timeout=30,
    ):
        self.ucert = ucert
        self.ukey = ukey
        self.passwd = passwd
        self.access_token = access_token
        self.verify = verify

        if capath:
            self.capath = capath
        elif "X509_CERT_DIR" in os.environ:
            self.capath = os.environ["X509_CERT_DIR"]
        else:
            self.capath = "/etc/grid-security/certificates"

        # Disable the warnings
        if not verify:
            urllib3.disable_warnings()

        self.connectTimeout = connectTimeout
        self.timeout = timeout

    def _handle_error(self, url, code, response_body=None):
        # Try parsing the response, maybe we can get the error message
        message = None
        response = None
        if response_body:
            try:
                response = json.loads(response_body)
                if "message" in response:
                    message = response["message"]
                else:
                    message = response_body
            except Exception:
                message = response_body

        if code == 207:
            try:
                raise ClientError("\n".join(map(lambda m: m["http_message"], response)))
            except (KeyError, TypeError):
                raise ClientError(message)
        elif code == 400:
            if message:
                raise ClientError("Bad request: " + message)
            else:
                raise ClientError("Bad request")
        elif 401 <= code <= 403:
            if message:
                raise Unauthorized(message)
            else:
                raise Unauthorized()
        elif code == 404:
            raise NotFound(url, message)
        elif code == 419:
            raise NeedDelegation("Need delegation")
        elif code == 424:
            raise FailedDependency("Failed dependency")
        elif 404 < code < 500:
            raise ClientError(str(code))
        elif code == 503:
            raise TryAgain(str(code))
        elif code >= 500:
            raise ServerError(str(code))

    def method(self, method, url, body=None, headers=None, user=None, passw=None):
        _headers = {"Accept": "application/json"}
        if headers:
            _headers.update(headers)
        if self.access_token:
            _headers["Authorization"] = "Bearer " + self.access_token

        auth = None
        if user and passw:
            from requests.auth import HTTPBasicAuth

            auth = HTTPBasicAuth(user, passw)

        if self.verify and self.capath:
            self.verify = self.capath

        response = requests.request(
            method=method,
            url=str(url),
            data=body,
            headers=_headers,
            verify=self.verify,
            timeout=(self.connectTimeout, self.timeout),
            cert=(self.ucert, self.ukey),
            auth=auth,
        )
        """
        verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``.
        """

        self._handle_error(url, response.status_code, response.text)

        return str(response.text)



#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json


class Submitter:
    def __init__(self, context):
        self.context = context

    @staticmethod
    def build_submission(
        transfers=None, delete=None, params=None, staging=None, **kwargs
    ):
        job = dict()
        job["params"] = dict()
        if params:
            job["params"].update(params)
        job["params"].update(kwargs)

        if delete:
            job["delete"] = delete
        if staging:
            job["staging"] = staging
        if transfers:
            job["files"] = transfers
            if "checksum" in kwargs:
                for f in job["files"]:
                    if "checksum" not in f:
                        f["checksum"] = job["params"]["checksum"]
                del job["params"]["checksum"]
            if "filesize" in job["params"]:
                for f in job["files"]:
                    if "filesize" not in f:
                        f["filesize"] = job["params"]["filesize"]
                del job["params"]["filesize"]
            if "file_metadata" in job["params"]:
                for f in job["files"]:
                    if "metadata" not in f:
                        f["metadata"] = job["params"]["file_metadata"]
                del job["params"]["file_metadata"]

        return json.dumps(job, indent=2)

    def submit(self, transfers=None, delete=None, params=None, **kwargs):
        job = Submitter.build_submission(
            transfers=transfers, delete=delete, params=params, **kwargs
        )
        r = json.loads(self.context.post_json("/jobs", job))
        return r["job_id"]

    def cancel(self, job_id, file_ids=None):
        if file_ids is not None:
            file_ids_str = ",".join(map(str, file_ids))
            return json.loads(
                self.context.delete("/jobs/%s/files/%s" % (job_id, file_ids_str))
            )
        else:
            return json.loads(self.context.delete("/jobs/%s" % job_id))

    def cancel_all(self, vo=None):
        if vo is None:
            return json.loads(self.context.delete("/jobs/all"))
        else:
            return json.loads(self.context.delete("/jobs/vo/%s" % vo))



#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
from urllib.parse import quote

from .exceptions import *


class Inquirer:
    def __init__(self, context):
        self.context = context

    def get_job_status(self, job_id, list_files=False):

        if not isinstance(job_id, str):
            raise Exception("The job_id provided is not a string!")

        try:
            job_info = json.loads(self.context.get("/jobs/%s" % job_id))

            if list_files:
                job_info["files"] = json.loads(
                    self.context.get("/jobs/%s/files" % job_id)
                )
                job_info["dm"] = json.loads(self.context.get("/jobs/%s/dm" % job_id))

            return job_info

        except NotFound:
            raise NotFound(job_id)

    def get_jobs_statuses(self, job_ids, list_files=False):

        if isinstance(job_ids, list):
            xfer_ids = ",".join(job_ids)
        else:
            raise Exception("The input provided is not a list of ids!")

        try:
            if not list_files:
                job_info = json.loads(self.context.get("/jobs/%s" % xfer_ids))
            else:
                job_info = json.loads(
                    self.context.get(
                        "/jobs/%s?files=file_state,dest_surl,finish_time,start_time,reason,source_surl,file_metadata"
                        % xfer_ids
                    )
                )

            return job_info
        except NotFound:
            raise NotFound(job_ids)

    def get_job_list(
        self,
        user_dn=None,
        vo_name=None,
        source_se=None,
        dest_se=None,
        delegation_id=None,
        state_in=None,
    ):
        url = "/jobs?"
        args = {}
        if user_dn:
            args["user_dn"] = user_dn
        if vo_name:
            args["vo_name"] = vo_name
        if source_se:
            args["source_se"] = source_se
        if dest_se:
            args["dest_se"] = dest_se
        if delegation_id:
            args["dlg_id"] = delegation_id
        if state_in:
            args["state_in"] = ",".join(state_in)

        query = "&".join("%s=%s" % (k, quote(v, "")) for k, v in args.items())
        url += query

        return json.loads(self.context.get(url))

    def whoami(self):
        return json.loads(self.context.get("/whoami"))



""""
SYSTEM REQUIREMENTS:
The following Centos packages are required for M2Crypto:
python3-devel openssl-devel swig gcc-c++
"""
from setuptools import setup

# This is required to allow editable pip installs while using the declarative configuration (setup.cfg)
setup()



#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


class FTS3ClientException(Exception):
    pass


class BadEndpoint(FTS3ClientException):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return "Bad endpoint: %s" % self.reason


class Unauthorized(FTS3ClientException):
    def __init__(self, reason=None):
        self.reason = reason

    def __str__(self):
        if self.reason:
            return "Unauthorized: %s" % self.reason
        else:
            return "Unauthorized"


class ClientError(FTS3ClientException):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return "Client error: %s" % self.reason


class NeedDelegation(ClientError):
    def __str__(self):
        return "Need to delegate credentials"


class FailedDependency(ClientError):
    def __str__(self):
        return "Failed dependency"


class ServerError(FTS3ClientException):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return "Server error: %s" % self.reason


class TryAgain(ServerError):
    def __str__(self):
        return "Try again: %s" % self.reason


class NotFound(FTS3ClientException):
    def __init__(self, resource, reason=None):
        self.resource = resource
        self.reason = reason

    def __str__(self):
        if self.reason:
            return self.reason
        else:
            return "Not found: %s" % self.resource



#   Copyright 2014-2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from fts3.rest.client import Inquirer


def list_jobs(
    context,
    user_dn=None,
    vo=None,
    source_se=None,
    dest_se=None,
    delegation_id=None,
    state_in=None,
):
    """
    List active jobs. Can filter by user_dn and vo

    Args:
        context:       fts3.rest.client.context.Context instance
        user_dn:       Filter by user dn. Can be left empty
        vo:            Filter by vo. Can be left empty
        delegation_id: Filter by delegation ID. Mandatory for state_in
        state_in:      Filter by job state. An iterable is expected (i.e. ['SUBMITTED', 'ACTIVE']

    Returns:
        Decoded JSON message returned by the server (list of jobs)
    """
    inquirer = Inquirer(context)
    return inquirer.get_job_list(
        user_dn, vo, source_se, dest_se, delegation_id, state_in
    )


def get_job_status(context, job_id, list_files=False):
    """
    Get a job status

    Args:
        context:    fts3.rest.client.context.Context instance
        job_id:     The job ID
        list_files: If True, the status of each individual file will be queried

    Returns:
        Decoded JSON message returned by the server (job status plus, optionally, list of files)
    """
    inquirer = Inquirer(context)
    return inquirer.get_job_status(job_id, list_files)


def get_jobs_statuses(context, job_ids, list_files=False):
    """
    Get status for a list of jobs

    Args:
        context:    fts3.rest.client.context.Context instance
        job_ids:    The job list
        list_files: If True, the status of each individual file will be queried

    Returns:
        Decoded JSON message returned by the server (job status plus, optionally, list of files)
    """
    inquirer = Inquirer(context)
    return inquirer.get_jobs_statuses(job_ids, list_files)



#   Copyright 2014-2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from fts3.rest.client import Inquirer


def whoami(context):
    """
    Queries the server to see how does it see us

    Args:
        context: fts3.rest.client.context.Context instance

    Returns:
        Decoded JSON message returned by the server with a representation of
        the user credentials (as set in context)
    """
    inquirer = Inquirer(context)
    return inquirer.whoami()



#   Copyright 2014-2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from fts3.rest.client import Ban


def ban_se(context, storage, status="cancel", timeout=0, allow_submit=False):
    """
    Ban a storage element

    Args:
        context: fts3.rest.client.context.Context instance
        storage: The storage to ban
        status:  The status of the banning: cancel or wait (leave queued jobs for some time)
        timeout: The wait timeout (0 means leave the queued jobs until they are done)
        allow_submit: If True, submissions will be accepted. Only meaningful if status=active

    Returns:
        List of job ids affected by the banning
    """
    ban = Ban(context)
    return ban.ban_se(storage, status, timeout, allow_submit)


def ban_dn(context, dn):
    """
    Ban a user

    Args:
        context: fts3.rest.client.context.Context instance
        dn:      The dn of the user to be banned

    Returns:
        List of job ids affected by the banning
    """
    ban = Ban(context)
    return ban.ban_dn(dn)


def unban_se(context, storage):
    """
    Unban a storage element

    Args:
        context: fts3.rest.client.context.Context instance
        storage: The storage to unban

    Returns:
        Nothing
    """
    ban = Ban(context)
    return ban.unban_se(storage)


def unban_dn(context, dn):
    """
    Unban a user

    Args:
        context: fts3.rest.client.context.Context instance
        dn:      The dn of the user to be unbanned

    Returns:
        Nothing
    """
    ban = Ban(context)
    ban.unban_dn(dn)



#   Copyright  Members of the EMI Collaboration, 2013.
#   Copyright 2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import timedelta

from .base import Base
from fts3.rest.client import Delegator as Deleg


class Delegator(Base):
    def __init__(self):
        super(Delegator, self).__init__(
            description="This command can be used to (re)delegate your credentials to the FTS3 server",
            example="""
            $ %(prog)s -s https://fts3-devel.cern.ch:8446
            Delegation id: 9a4257f435fa2010"
            """,
        )

        self.opt_parser.add_option(
            "-f",
            "--force",
            dest="force",
            default=False,
            action="store_true",
            help="force the delegation",
        )

        self.opt_parser.add_option(
            "-H",
            "--hours",
            dest="duration",
            default=12,
            type="int",
            help="Duration of the delegation in hours (Default: 12)",
        )

    def run(self):
        context = self._create_context()
        delegator = Deleg(context)
        delegation_id = delegator.delegate(
            lifetime=timedelta(hours=self.options.duration), force=self.options.force
        )
        self.logger.info("Delegation id: %s" % delegation_id)
        self.logger.debug(
            "Termination time: %s" % delegator.get_info()["termination_time"]
        )
        return delegation_id



#   Copyright 2014-2020 CERN
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import timedelta
from fts3.rest.client import Delegator


def delegate(
    context,
    lifetime=timedelta(hours=7),
    force=False,
    delegate_when_lifetime_lt=timedelta(hours=2),
):
    """
    Delegates the credentials

    Args:
        context:  fts3.rest.client.context.Context instance
        lifetime: The delegation life time
        force:    If true, credentials will be re-delegated regardless
                  of the remaining life of the previous delegation
        delegate_when_lifetime_lt: If the remaining lifetime on the delegated proxy is less than this interval,
                  do a new delegation

    Returns:
        The delegation ID
    """
    delegator = Delegator(context)
    return delegator.delegate(lifetime, force, delegate_when_lifetime_lt)
